from bge import logic, render
from io import StringIO
from uplogic.ui.canvas import Canvas
from uplogic.ui.layout import RelativeLayout
from uplogic.ui.label import Label
from uplogic.ui.textinput import TextInput
from uplogic.data import GlobalDB
from uplogic.utils.raycasting import raycast_screen
from uplogic.utils.scene import world_to_screen
from uplogic.utils.math import cycle
from uplogic.utils.math import clamp
from uplogic.input.keyboard import key_down, key_pulse
from uplogic.input.mouse import mouse_down, mouse_wheel
import bpy, blf
import bge
import sys, os
from datetime import datetime
from mathutils import Vector
import uplogic


GLOBALS = {}


def set_global(key, val):
    global GLOBALS
    GLOBALS[key] = val


def _get_globals():
    scene = logic.getCurrentScene()
    global GLOBALS
    GLOBALS['scene'] = scene
    GLOBALS['bpy'] = bpy
    GLOBALS['bge'] = bge
    GLOBALS['uplogic'] = uplogic
    GLOBALS['logic'] = logic
    GLOBALS['render'] = render
    GLOBALS['console'] = get_console(True)
    for obj in scene.objects:
        GLOBALS[obj.blenderObject.name] = obj
    return GLOBALS


def enable(toggle_key='F12', visible=False):
    scene = logic.getCurrentScene()
    c = get_console(True, toggle_key=toggle_key, visible=visible)
    c.scene = scene
    if disable not in scene.onRemove:
        scene.onRemove.append(disable)
    if c.update not in scene.pre_draw:
        c.register()
        scene.pre_draw.append(c.update)
    sys.stdout = Console()


def disable():
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__
    console = get_console(True)
    if console:
        console.stop()
    consoles = GlobalDB.retrieve('uplogic.consoles')
    consoles.remove('default')


class Console(StringIO):

    def write(self, __s: str) -> int:
        _print(__s)


class ErrorConsole(StringIO):
    def write(self, __s: str) -> int:
        error(__s)

COLORS = {
    'LOG': [1, 1, 1, 1],
    'INFO': [.3, .8, 1, 1],
    'DEBUG': [1, 1, .6, 1],
    'WARNING': [1, .8, .2, 1],
    'ERROR': [1, .3, .3, 1],
    'CRITICAL': [1, .0, .0, 1],
    'SUCCESS': [.3, 1, .3, 1]
}


class CommandLabel(Label):
    pass


def set_log_level(level: int = 2) -> None:
    """Set the log level for the console. Higher values mean less important messages will be hidden.

    - 0: No Restrictions
    - 1: Debug
    - 2: Info (Default)
    - 3: Warning
    - 4: Error
    - 5: Critical

    Args:
        level (int): Log Level
    """
    get_console().log_level = clamp(level, 0, 5)


def set_log_directory(path):
    get_console().log_directory = path


class ConsoleLayout(Canvas):
    opacity = 1
    padding = [5, 10]
    toggle_key = 'F12'

    
    @property
    def log_directory(self):
        """Path to the log folder. This should be a directory, not a file."""
        return self._log_directory

    @log_directory.setter
    def log_directory(self, folder):
        if folder is not None:
            if not os.path.isdir(folder):
                os.makedirs(folder)
            self._log_file = os.path.join(
                folder,
                f'log.{len(os.listdir(folder))}.txt'
            )
        else:
            self._log_file = None
        self._log_directory = folder

    @property
    def log_file(self):
        return self._log_file

    @log_file.setter
    def log_file(self, val):
        error("'ConsoleLayout.log_file' is read-only!")
    
    def __init__(self, toggle_key='F12', visible=False): 
        self.log_directory = None
        scene = logic.getCurrentScene()
        log_levels = {
            'NOTSET': 0,
            'DEBUG': 1,
            'INFO': 2,
            'WARNING': 3,
            'ERROR': 4,
            'CRITICAL': 5
        }
        self.log_level = log_levels.get(bpy.data.scenes[scene.name].game_settings.log_level)
        self.toggle_key = toggle_key
        self._mouse_visible = False
        self.issued_commands = []
        self._mouse_down = False
        self._goback_index = -1
        super().__init__()
        self.input = TextInput(text='', shadow=True, valign='center')
        self.input.on_enter = self.on_enter
        self.input.edit = True
        self.console = RelativeLayout(relative={'size': True, 'pos': True}, pos=[0, 0], size=(1, .4))
        self._layout = self.console
        self.layout = RelativeLayout(relative={'size': True}, size=(1, 1), bg_color=[0, 0, 0, .4])
        self.layout.use_clipping = True
        self.add_widget(self.console)
        self.console.add_widget(self.layout)
        self.layout.add_widget(self.input)
        self.fade_event = None
        self._toggle_key = False
        self._prev_msg = None
        if disable not in scene.onRemove:
            scene.onRemove.append(disable)
        if self.update not in scene.pre_draw:
            scene.pre_draw.append(self.update)
        self.nameplate = Label(text='', shadow=True, relative={'pos': True}, halign='center', font_size=13)
        self.nameplate.update = self.update_nameplate
        self.canvas.add_widget(self.nameplate)
        self.font_size = 12
        self.position = 'bottom'
        self.info_mode = getattr(bpy.context.scene, 'screen_console_open', False)
        self.active = False

    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, val):
        for w in self._layout.children:
            val.add_widget(w)
        self._layout = val

    @property
    def info_mode(self):
        return self._info_mode

    @info_mode.setter
    def info_mode(self, val):
        self._info_mode = val
        self.active = self.active

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, val):
        if self.info_mode:
            self.show = True
            self.layout.bg_color = (0, 0, 0, .4) if val else (0, 0, 0, 0)
            self.input.show = val
            self.nameplate.show = val
        else:
            self.show = val
        self._active = val

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, val):
        if val == 'center':
            self.console.valign = val
            self.console.halign = val
            self.console.pos = (.5, .5)
            self.console.size = (.4, .4)
            self._position = val
        elif val == 'left':
            self.console.valign = 'bottom'
            self.console.halign = val
            self.console.size = (.4, 1)
            self.console.pos = (0, 0)
            self._position = val
        elif val == 'right':
            self.console.valign = 'bottom'
            self.console.halign = val
            self.console.pos = (1, 0)
            self.console.size = (.4, 1)
            self._position = val
        elif val == 'top':
            self.console.valign = 'top'
            self.console.halign = 'left'
            self.console.pos = (0, 1)
            self.console.size = (1, .4)
            self._position = val
        elif val == 'bottom':
            self.console.valign = 'bottom'
            self.console.halign = 'left'
            self.console.pos = (0, 0)
            self.console.size = (1, .4)
            self._position = val
        else:
            error(f'"{val}" not recognized.')
        self.arrange()

    @property
    def font_size(self):
        return self._font_size

    @font_size.setter
    def font_size(self, val):
        val = clamp(val, 5, 30)
        self._font_size = int(val)
        for c in self.layout.children:
            c.font_size = int(val)
        self.nameplate.font_size = int(val)
        self.arrange()

    def on_enter(self):
        if self.input.text:
            self.issued_commands.append(self.input.text)
        sys.__stdout__.write(f'>{self.input.text}\n')
        self.add_message(f'>{self.input.text}')
        command_name = self.input.text.split(' ')[0]
        command = Commands.commands.get(command_name)
        if command:
            command.invoke(self.input.text)
        else:
            try:
                exec(self.input.text, _get_globals())
            except Exception as e:
                error(f'Error occured when executing command "{command_name}":\n{e}')
        self._goback_index = -1
        self.input.text = ''
        self.input.edit = True

    def update_nameplate(self):
        ray = raycast_screen()
        mdown = mouse_down()
        if ray.obj:
            self.nameplate.pos = world_to_screen(ray.point) + Vector((0, .01))
            self.nameplate.text = ray.obj.blenderObject.name
            if mdown and not self._mouse_down:
                self.input.write(self.nameplate.text)
        else:
            self.nameplate.text = ''
        self._mouse_down = mdown

    def toggle(self):
        if not self.show:
            self._mouse_visible = logic.mouse.visible
        if not self._toggle_key:
            self.active = not self.active
            self.opacity = 1
            self._toggle_key = True

    def update(self):
        if self.input.edit != self.show:
            self.input.edit = self.show
        move_goback = key_pulse('UPARROW') - key_pulse('DOWNARROW')
        if key_down(self.toggle_key):
            self.toggle()
        elif not self.show:
            self._toggle_key = False
            return
        elif move_goback:
            if not self.issued_commands:
                pass
            elif not self._toggle_key:
                self._goback_index = clamp(self._goback_index + move_goback, -1, len(self.issued_commands) - 1)
                if self._goback_index < 0:
                    self.input.text = ''
                else:
                    self.input.text = list(self.issued_commands.__reversed__())[self._goback_index]
                self.input.move_cursor_to_end()
            self._toggle_key = True
        else:
            self._toggle_key = False
        logic.mouse.visible = self.active or self._mouse_visible

    def stop(self):
        self.clear()
        scene = logic.getCurrentScene()
        if self.toggle in scene.pre_draw:
            scene.pre_draw.remove(self.toggle)

    def add_message(self, msg, type='LOG', time=True, command=False):
        if (msg == ' ' or self._prev_msg == ' ') and len(self.layout.children):
            self.layout.children[-1].text += msg
            self._prev_msg = msg
            return
        now = datetime.now()
        current_time = f'[{now.strftime("%H:%M:%S")}]' if time else "\t\t\t\t  ".replace('\t', '    ')
        label_class = CommandLabel if command else Label
        self.layout.add_widget(label_class(text=f'{current_time}  {msg}', pos=[5, 10], font_color=COLORS[type], shadow=True, font_size=self.font_size))
        self._prev_msg = msg
        self.arrange()

    def arrange(self):
        blf.size(0, self.font_size)
        dim = blf.dimensions(0, 'A')
        cheight = dim[1]
        cwidth = dim[0]
        lheight = self.layout._draw_size[1]
        amount = lheight / cheight
        y = cheight * 2.4 * 1.5
        for i, child in enumerate(self.layout._children_reversed):
            if child is self.input:
                continue
            child.pos[1] = y
            y += cheight * 1.5
            if child.pos[1] > lheight - cheight:
                self.layout.remove_widget(child)
            child.opacity = 1 - (i * (1/amount))
            child.shadow_color[3] = child.font_color[3]
        self.input.pos[1] = self.font_size + 3
        self.input.pos[0] = cwidth


def get_console(create=False, toggle_key='F12', visible=False) -> ConsoleLayout:
    consoles = GlobalDB.retrieve('uplogic.consoles')
    console = consoles.get('default')
    if console is None and create:
        console = ConsoleLayout(toggle_key=toggle_key, visible=visible)
        consoles.put('default', console)
        console.add_message('On-Screen Console active; Check System Console for Errors.', type='DEBUG')
    return console


class ansicol:
    RED = '\033[31m\033[1m'
    GREEN = '\033[32m\033[1m'
    YELLOW = '\033[33m'
    BYELLOW = '\033[93m'
    BBLUE = '\033[36m'
    END = '\033[0m'


def write(*msg, type='LOG'):
    msg = ' '.join([m.__repr__() if not isinstance(m, str) else m for m in msg])
    _f = {
        'LOG': log,
        'INFO': info,
        'DEBUG': debug,
        'WARNING': warning,
        'ERROR': error,
        'SUCCESS': success
    }

    _f[type](msg)


def _create_msg(msg, log_lvl, type: str, color):
    console = get_console(True)
    if console is None:
        print(sysmsg)
        return
    if console.log_level > log_lvl:
        return

    msg = ' '.join([m.__repr__() if not isinstance(m, str) else m for m in msg])
    msg += '\n'
    sysmsg = f'{color}{type}{ansicol.END}: {msg}'
    show_time = True
    sys.__stdout__.write(f'{sysmsg}')
    for msg in str(msg).split('\n'):
        if msg:
            msg.replace('  ', '    ')
            console.add_message(f'{msg}', type.upper(), time=show_time)
            show_time = False
            if console.log_file is not None:
                with open(console.log_file, 'a') as f:
                    date = datetime.today().strftime('%Y-%m-%d %H:%M:%S')
                    f.write(f'[{date}][{type}]\t{msg}\n')


def _print(msg):
    console = get_console(True)
    if console is None:
        print(sysmsg)
        return

    msg = ''.join([m.__repr__() if not isinstance(m, str) else m for m in msg])
    sysmsg = msg
    show_time = True
    sys.__stdout__.write(f'{sysmsg}')
    for msg in str(msg).split('\n'):
        if msg:
            msg.replace('  ', '    ')
            console.add_message(f'{msg}', 'LOG', time=show_time)
            show_time = False
            if console.log_file is not None:
                with open(console.log_file, 'a') as f:
                    date = datetime.today().strftime('%Y-%m-%d %H:%M:%S')
                    f.write(f'[{date}]\t{msg}\n')


def log(*msg, type='LOG'):
    """Write to the console in a generic fashion.

    If `ConsoleLayout.log_directory` is defined, the message will be added to the log file.

    Args:
        type (str, optional): Message type of ['LOG', 'SUCCESS', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']. Defaults to 'LOG'.
    """
    _create_msg(msg, 5, type, '')


def success(*msg):
    """Write to the console with green colors if the `ConsoleLayout.log_level` is 0.

    If `ConsoleLayout.log_directory` is defined, the message will be added to the log file.
    """
    _create_msg(msg, 0, 'Success', ansicol.GREEN)


def debug(*msg):
    """Write to the console with bright yellow colors if the `ConsoleLayout.log_level` is 1 or less.

    If `ConsoleLayout.log_directory` is defined, the message will be added to the log file.
    """
    _create_msg(msg, 1, 'Debug', ansicol.BYELLOW)


def info(*msg):
    """Write to the console with blue colors if the `ConsoleLayout.log_level` is 2 or less.

    If `ConsoleLayout.log_directory` is defined, the message will be added to the log file.
    """
    _create_msg(msg, 2, 'Info', ansicol.BBLUE)


def warning(*msg):
    """Write to the console with yellow colors if the `ConsoleLayout.log_level` is 3 or less.

    If `ConsoleLayout.log_directory` is defined, the message will be added to the log file.
    """
    _create_msg(msg, 3, 'Warning', ansicol.YELLOW)


def error(*msg):
    """Write to the console with red colors if the `ConsoleLayout.log_level` is 4 or less.

    If `ConsoleLayout.log_directory` is defined, the message will be added to the log file.
    """
    _create_msg(msg, 4, 'Error', ansicol.RED)


def critical(*msg):
    """Write to the console with red colors if the `ConsoleLayout.log_level` is 5 or less.

    If `ConsoleLayout.log_directory` is defined, the message will be added to the log file.
    """
    _create_msg(msg, 5, 'Critical', ansicol.RED)


nodeprefs = bpy.context.preferences.addons.get('bge_netlogic', None)
if nodeprefs and getattr(bpy.context.scene, 'use_screen_console', True):
    enable(toggle_key='F12')


##############################################################
# Predefined Commands
##############################################################


class Commands:
    commands = {}

    @classmethod
    def add_command(cls, command):
        cls.commands[command.command] = command


def add_command(command):
    Commands.add_command(command)
    return command


def console_command(command):
    Commands.add_command(command)
    return command

class Command:
    """
    Command to be executed via the On-Screen Console.

    Set the following attributes on subclass:
    - `command`: id of the command
    - `usage`: sequence of argument names (for a command "setres 1920 1080", put "RES_X RES_Y")
    - `arg_count` (default 0): number of possible arguments
    - `description`: Shown when executing "help -d" command

    Also, override 
    ```
    @classmethod
    def execute(cls, args):
        ...
    ```

    If decorated with `@console_command`, this command will be added to the Console when imported.
    """
    command = ''
    usage = ''
    arg_count = 0
    description = ''

    @classmethod
    def invoke(cls, message):
        args = message.split(' ')
        args = args[1:]
        if len(args) < cls.arg_count:
            debug(f'Usage: "{cls.command} {cls.usage}"')
            return
        try:
            cls.execute(args)
        except Exception as e:
            error(e)

    @classmethod
    def execute(cls, args):
        pass


@console_command
class RemoveObjectCommand(Command):
    command = 'remove'
    arg_count = 1
    usage = 'OBJECT_ID'
    description = 'Remove an object from the current scene.'

    @classmethod
    def execute(cls, args):
        object_name = args[0]
        scene = logic.getCurrentScene()
        scene.getGameObjectFromObject(
            bpy.data.objects[object_name]
        ).endObject()


@console_command
class DisableCommand(Command):
    command = 'disable'
    arg_count = 1
    usage = 'OBJECT_ID'
    description = 'Set an object to invisible and inactive.'

    @classmethod
    def execute(cls, args):
        object_name = args[0]
        scene = logic.getCurrentScene()
        obj = scene.getGameObjectFromObject(
            bpy.data.objects[object_name]
        )
        obj.setVisible(False)
        obj.suspendPhysics()
        obj.suspendDynamics()


@console_command
class EnableCommand(Command):
    command = 'enable'
    arg_count = 1
    usage = 'OBJECT_ID'
    description = 'Set an object to visible and active.'

    @classmethod
    def execute(cls, args):
        object_name = args[0]
        scene = logic.getCurrentScene()
        obj = scene.getGameObjectFromObject(
            bpy.data.objects[object_name]
        )
        obj.setVisible(True)
        obj.restorePhysics()
        obj.restoreDynamics()


@console_command
class ShowInfoCommand(Command):
    command = 'showinfo'
    arg_count = 1
    usage = 'STAGE(0-3)'
    description = 'Show game info. 1: Show FPS, 2: Show Profile, 3: Show Properties'

    @classmethod
    def execute(cls, args):
        stage = args[0]
        render.showFramerate(int(stage) > 0)
        render.showProfile(int(stage) > 1)
        render.showProperties(int(stage) > 2)


@console_command
class QuitCommand(Command):
    command = 'quit'
    arg_count = 0
    usage = ''
    description = 'Quit the game.'

    @classmethod
    def execute(cls, args):
        logic.endGame()


@console_command
class RestartCommand(Command):
    command = 'restart'
    arg_count = 0
    usage = ''
    description = 'Restart the game.'

    @classmethod
    def execute(cls, args):
        logic.restartGame()


@console_command
class PrintCommand(Command):
    command = 'print'
    arg_count = 1
    usage = 'MESSAGE'
    description = 'Print a message or variable.'

    @classmethod
    def execute(cls, args):
        msg = args[0]
        print(eval(msg, _get_globals()))


# @console_command
# class PauseCommand(Command):
#     command = 'pause'
#     arg_count = 1
#     usage = 'STATE(0 or 1)'

#     @classmethod
#     def execute(cls, args):
#         state = args[0]
#         if state:
#             logic.getCurrentScene().suspend()
#         else:
#             logic.getCurrentScene().resume()


@console_command
class HelpCommand(Command):
    command = 'help'
    description = 'Print out all available commands.'

    @classmethod
    def execute(cls, args):
        mode = None
        msg = 'Available Commands:'
        if len(args) > 0:
            mode = args[0]
        for command in Commands.commands.values():
            if mode == '-d':
                msg += f'\n -  "{command.command} {command.usage}"    -    {command.description}'
            else:
                msg += f'\n -  "{command.command} {command.usage}"'
        print(msg)


@console_command
class FontSizeCommand(Command):
    command = 'fontsize'
    arg_count = 1
    usage = 'PX_SIZE'
    description = 'Modify the font size of the console.'

    @classmethod
    def execute(cls, args):
        size = args[0]
        get_console(True).font_size = int(size)


@console_command
class SetResolutionCommand(Command):
    command = 'setres'
    arg_count = 2
    usage = 'RES_X RES_Y'
    description = 'Set the app resolution'

    @classmethod
    def execute(cls, args):
        rx, ry = args
        render.setWindowSize(int(rx), int(ry))


@console_command
class ToggleDebugCommand(Command):
    command = 'debug'
    arg_count = 0
    description = 'Toggle Debug Mode'

    @classmethod
    def execute(cls, args):
        console = get_console(True)
        console.info_mode = not console.info_mode
        debug(f'Console Debug Mode is now {"ON" if console.info_mode else "OFF"}.')



from uplogic.nodes import ULActionNode


class ULAlignAxisToVector(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.vector = None
        self.axis = None
        self.factor = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_object = self.get_input(self.game_object)
        v = self.get_input(self.vector).copy()
        axis = self.get_input(self.axis)
        factor = self.get_input(self.factor)
        if not self.local:
            v -= game_object.worldPosition
        if axis > 2:
            v.negate()
            axis -= 3
        game_object.alignAxisToVect(v, axis, factor)
        self._done = True



from uplogic.nodes import ULParameterNode
from uplogic.utils.math import rotate2d
from uplogic.utils.math import rotate3d
from uplogic.utils.math import rotate_by_axis
from uplogic.utils.math import rotate_by_euler
from mathutils import Vector
from math import degrees


class ULRotateByPoint(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        # self.mode = 0
        self.global_axis = 0
        self.origin = None
        self.pivot = None
        self.angle = None
        self.arbitrary_axis = None
        self.euler_angles = None
        self.OUT = self.add_output(self.get_point)
        self._operations = [
            self._rotate2d,
            self._rotate3d,
            self._rotate_by_axis,
            self._rotate_by_euler
        ]

    @property
    def mode(self):
        return None

    @mode.setter
    def mode(self, val):
        self.OUT._value_getter = self._operations[val]

    def _rotate2d(self):
        return rotate2d(
            self.get_input(self.origin),
            self.get_input(self.pivot),
            degrees(self.get_input(self.angle))
        )

    def _rotate3d(self):
        return rotate3d(
            self.get_input(self.origin),
            self.get_input(self.pivot),
            degrees(self.get_input(self.angle)),
            self.global_axis
        )

    def _rotate_by_axis(self):
        return rotate_by_axis(
            self.get_input(self.origin),
            self.get_input(self.pivot),
            degrees(self.get_input(self.angle)),
            self.get_input(self.arbitrary_axis)
        )

    def _rotate_by_euler(self):
        return rotate_by_euler(
            self.get_input(self.origin),
            self.get_input(self.pivot),
            self.get_input(self.euler_angles)
        )

    def get_point(self):
        return Vector((0, 0, 0))



from uplogic.nodes import ULActionNode
from uplogic.utils.raycasting import raycast
from uplogic.utils.raycasting import raycast_screen
from uplogic.utils.raycasting import raycast_projectile
from uplogic.utils.raycasting import raycast_mouse
from uplogic.utils.raycasting import RayCastData
from uplogic.utils.math import get_bitmask
from bge.logic import getCurrentScene
from bpy.types import Material
from bge.types import KX_GameObject
from mathutils import Vector


class ULRaycast(ULActionNode):  # XXX: Deprecated!

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = False
        self.caster = None
        self.origin = None
        self.destination = None
        self.local: bool = None
        self.property_name: str = None
        self.material: str = None
        self.exclude: str = None
        self.xray: bool = None
        self.custom_dist: bool = None
        self.distance: float = None
        self.mask: int = get_bitmask(all=True)
        self.visualize: bool = None
        self._data = RayCastData((None, None, None, None, None, None))
        self.face_data = False
        self.RESULT = self.add_output(self.get_result)
        self.PICKED_OBJECT = self.add_output(self.get_picked_object)
        self.POINT = self.add_output(self.get_point)
        self.NORMAL = self.add_output(self.get_normal)
        self.DIRECTION = self.add_output(self.get_direction)
        self.MATERIAL = self.add_output(self.get_material)
        self.UV = self.add_output(self.get_uv)
        self.network = None

    def setup(self, network):
        self.network = network

    def get_result(self):
        return self._data.obj is not None

    def get_picked_object(self):
        return self._data.obj

    def get_point(self):
        return self._data.point

    def get_normal(self):
        return self._data.normal

    def get_direction(self):
        return self._data.direction

    def get_material(self):
        return self._data.face.material if self._data.face else None

    def get_uv(self):
        return self._data.uv

    def evaluate(self):
        condition = self.get_condition()
        if not condition:
            self._data = RayCastData((None, None, None, None, None, None))
            return
        origin = self.get_input(self.origin)
        caster = self.get_input(self.caster)
        destination = self.get_input(self.destination)
        local: bool = self.get_input(self.local)
        property_name: str = self.get_input(self.property_name)
        material: Material = self.get_input(self.material)
        # exclude: str = self.get_input(self.exclude)
        xray: bool = self.get_input(self.xray)
        distance: float = self.get_input(self.distance)
        visualize: bool = self.get_input(self.visualize)
        self._data = raycast(
            caster,
            origin,
            destination,
            distance=distance if self.get_input(self.custom_dist) else 0,
            prop=property_name,
            material=material.name if material else '',
            # exclude=exclude,
            xray=xray,
            local=local,
            mask=self.get_input(self.mask),
            visualize=visualize,
            face_data=self.face_data
        )


class RaycastNode(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.mode = 0
        self.condition: bool = False
        self.caster: KX_GameObject = None
        self.camera: KX_GameObject = None
        self.origin: Vector = Vector((0, 0, 0))
        self.target: Vector = Vector((0, 0, 1))
        self.aim: Vector = Vector((.5, .5))
        self.power: float = 10
        self.resolution: float = .9
        self.local: bool = False
        self.property_name: str = None
        self.material: str = None
        self.xray: bool = False
        self.use_custom_distance: bool = False
        self.distance: float = 0.0
        self.mask: int = get_bitmask(all=True)
        self.use_custom_gravity: bool = False
        self.gravity: Vector = None
        self.visualize: bool = False
        self._data = RayCastData((None, None, None, None, None, None))
        self.face_data = False
        self.RESULT = self.add_output(self.get_result)
        self.PICKED_OBJECT = self.add_output(self.get_picked_object)
        self.POINT = self.add_output(self.get_point)
        self.NORMAL = self.add_output(self.get_normal)
        self.DIRECTION = self.add_output(self.get_direction)
        self.MATERIAL = self.add_output(self.get_material)
        self.UV = self.add_output(self.get_uv)
        self.PARABOLA = self.add_output(self.get_parabola)

    def get_result(self):
        return self._data.obj is not None

    def get_picked_object(self):
        return self._data.obj

    def get_point(self):
        return self._data.point

    def get_normal(self):
        return self._data.normal

    def get_direction(self):
        return self._data.direction

    def get_material(self):
        return self._data.face.material if self._data.face else None

    def get_uv(self):
        return self._data.uv

    def get_parabola(self):
        return self._data.points

    def evaluate(self):
        condition = self.get_condition()
        if not condition:
            self._data = RayCastData((None, None, None, None, None, None))
            return
        mode = self.mode

        material = self.get_input(self.material)

        if mode == 0:
            self._data = raycast(
                self.get_input(self.caster),
                self.get_input(self.origin),
                self.get_input(self.target),
                distance=self.get_input(self.distance) if self.get_input(self.use_custom_distance) else 0,
                prop=self.get_input(self.property_name),
                material=material.name if material else '',
                xray=self.get_input(self.xray),
                local=self.get_input(self.local),
                mask=self.get_input(self.mask),
                face_data=self.face_data,
                visualize=self.get_input(self.visualize)
            )
        elif mode == 1:
            self._data = raycast_projectile(
                self.get_input(self.caster),
                self.get_input(self.origin),
                self.get_input(self.target),
                power=self.get_input(self.power),
                distance=self.get_input(self.distance),
                resolution=1-self.get_input(self.resolution),
                prop=self.get_input(self.property_name),
                material=material.name if material else '',
                xray=self.get_input(self.xray),
                local=self.get_input(self.local),
                mask=self.get_input(self.mask),
                gravity=self.get_input(self.gravity) if self.get_input(self.use_custom_gravity) else None,
                face_data=self.face_data,
                visualize=self.get_input(self.visualize)
            )
        else:
            self._data = raycast_screen(
                self.get_input(self.caster),
                self.get_input(self.aim) if mode == 3 else None,
                distance=self.get_input(self.distance),
                prop=self.get_input(self.property_name),
                material=material.name if material else '',
                xray=self.get_input(self.xray),
                mask=self.get_input(self.mask),
                face_data=self.face_data
            )



from .shader import Filter2D


glsl = """
uniform sampler2D bgl_RenderedTexture;

// in vec4 bgl_TexCoord;
uniform float power;
uniform float factor;

// out vec4 fragColor;

void main()
{
    float pow = clamp(power, 0.0, 1.0);
	vec4 px =  texture(bgl_RenderedTexture, bgl_TexCoord.xy);
    bool cutoff = (bgl_TexCoord.y > 1.0-factor || bgl_TexCoord.y < factor);

    fragColor = mix(
        px,
        vec4(0.0, 0.0, 0.0, 1.0),
        (cutoff ? 1.0 : 0.0) * pow
    );
}
"""


class Letterbox(Filter2D):

    def __init__(self, power: float = 1.0, factor: float = 0.1, idx: int = None) -> None:
        self.uniforms = {'power': float(power), 'factor': float(factor)}
        super().__init__(glsl, idx, {'power': self.uniforms, 'factor': self.uniforms})

    @property
    def power(self):
        return self.uniforms['power']

    @power.setter
    def power(self, val):
        self.uniforms['power'] = float(val)
    
    @property
    def factor(self):
        return self.uniforms['factor']

    @factor.setter
    def factor(self, val):
        self.uniforms['factor'] = float(val)



from uplogic.events import receive
from uplogic.events import Event
from uplogic.nodes import ULConditionNode


class ULHandleEvent(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.subject = None
        self.event = None
        self.target = None
        self.OUT = self.add_output(self.get_received)
        self.BODY = self.add_output(self.get_body)
        self.TARGET = self.add_output(self.get_target)

    def get_received(self):
        return isinstance(self.event, Event)

    def get_body(self):
        return self.event.content if self.event else None

    def get_target(self):
        return self.event.messenger if self.event else None

    def evaluate(self):
        target = self.get_input(self.target)
        self.event = receive(self.get_input(self.subject), target if target else self.network.owner)



from uplogic.nodes import ULActionNode
from uplogic.utils.pooling import SpawnPool, Spawn, SimpleBullet, PhysicsBullet, SpawnedInstance
from uplogic.events import send, receive
from uplogic.utils import get_bitmask
from bge.types import KX_GameObject


class ULSpawnPool(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.spawn = None
        self.spawner = None
        self.amount = None
        self.object_instance = None
        self.life = None
        self.speed = None
        self.bitmask = get_bitmask(all=True)
        self.visualize = None
        self.create_on_init = True
        self.spawn_type = 'Simple'
        self._hit_evt = None
        self._spawned = False

        self._pool = None

        self.OUT = self.add_output(self.get_done)
        self.SPAWNED = self.add_output(self.get_spawned)
        self.ONHIT = self.add_output(self.get_hit)
        self.HITOBJECT = self.add_output(self.get_hit_obj)
        self.HITPOINT = self.add_output(self.get_hit_point)
        self.HITNORMAL = self.add_output(self.get_hit_normal)
        self.HITDIR = self.add_output(self.get_hit_direction)

        class NodeSimple(Spawn):
            pass

        class NodeSimpleBullet(SimpleBullet):
            def hit(self, data):
                send(self._pool, data)

        class NodePhysicsBullet(PhysicsBullet):
            def hit(self, data):
                send(self._pool, data)

        self.spawn_types = {
            'Simple': NodeSimple,
            'SimpleBullet': NodeSimpleBullet,
            'PhysicsBullet': NodePhysicsBullet,
            0: NodeSimple,
            1: NodeSimpleBullet,
            2: NodePhysicsBullet,
            3: SpawnedInstance
        }

    def get_done(self):
        return self._done

    def get_spawned(self):
        r = self._spawned
        self._spawned = False
        return r

    def get_hit(self):
        return self._hit_evt is not None

    def get_hit_obj(self):
        if self._hit_evt:
            return self._hit_evt.content.obj

    def get_hit_point(self):
        if self._hit_evt:
            return self._hit_evt.content.point

    def get_hit_normal(self):
        if self._hit_evt:
            return self._hit_evt.content.normal

    def get_hit_direction(self):
        if self._hit_evt:
            return self._hit_evt.content.direction

    def setup(self, tree):
        if self.create_on_init:
            inst = self.get_input(self.object_instance)
            self._pool = SpawnPool(
                object_name=getattr(inst, 'name', inst),
                amount=self.get_input(self.amount),
                lifetime=self.get_input(self.life),
                spawner=self.get_input(self.spawner),
                spawn=self.spawn_types[self.get_input(self.spawn_type)],
                raycast_mask=self.get_input(self.bitmask),
                visualize=self.get_input(self.visualize)
            )
        self._done = True

    def evaluate(self):
        spawn = self.get_input(self.spawn)
        if self._pool:
            self._hit_evt = receive(self._pool)
        if spawn and self._pool:
            self._pool.spawn()
            self.spawn_types[self.spawn_type].speed = self.get_input(self.speed)
            self._spawned = True

        if self.get_condition():
            inst: KX_GameObject = self.get_input(self.object_instance)
            self._pool = SpawnPool(
                getattr(inst, 'name', inst),
                self.get_input(self.amount),
                self.get_input(self.life),
                self.get_input(self.spawner),
                self.spawn_types[self.get_input(self.spawn_type)],
                visualize=self.get_input(self.visualize)
            )
            self._done = True



from uplogic.nodes import ULActionNode


class ULApplyRotation(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.rotation = None
        self.local = False
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_object = self.get_input(self.game_object)
        rotation = self.get_input(self.rotation)
        local = self.local
        if rotation:
            if len(rotation) == 3:
                game_object.applyRotation(rotation, local)
            else:
                game_object.applyRotation(rotation.to_euler("XYZ"), local)
        self._done = True



from uplogic.nodes import ULConditionNode


class ULAnd(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.ca = None
        self.cb = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        ca = self.get_input(self.ca)
        return ca and self.get_input(self.cb)


class ULAndList(ULConditionNode):

    def __init__(self):
        ULConditionNode.__init__(self)
        self.ca = True
        self.cb = True
        self.cc = True
        self.cd = True
        self.ce = True
        self.cf = True
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        results = [
            self.get_input(self.ca),
            self.get_input(self.ca),
            self.get_input(self.ca),
            self.get_input(self.ca),
            self.get_input(self.ca),
            self.get_input(self.ca)
        ]
        return False not in results



from uplogic.nodes import ULActionNode
import bpy
import json
import os
from uplogic import console


class ULListVariables(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.print_list = None
        self.file_name = None
        self.path = ''
        self.items = None
        self.OUT = self.add_output(self.get_done)
        self.LIST = self.add_output(self.get_list)

    def get_done(self):
        return self._done

    def get_list(self):
        return self.items

    def write_to_json(self, path, p_l):
        data = None
        if not path.endswith('.json'):
            path = os.path.join(path, f'{self.get_input(self.file_name)}.json')
        if os.path.isfile(path):
            f = open(path, 'r')
            data = json.load(f)
            if len(data) == 0:
                console.debug('There are no saved variables')
                return
            li = []
            for x in data:
                if p_l:
                    console.log('{}\t->\t{}'.format(x, data[x]))
                li.append(x)
            self.items = li
        else:
            console.debug('There are no saved variables')
        f.close()

    def evaluate(self):
        if not self.get_condition():
            return
        print_list = self.get_input(self.print_list)
        path = self.get_input(self.path)

        path = (bpy.path.abspath(path))
        os.makedirs(path, exist_ok=True)

        self.write_to_json(path, print_list)
        self._done = True



from uplogic.audio import Sound2D
from uplogic.nodes import ULActionNode
from bpy.types import Sound


class ULStartSound(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.sound = None
        self.loop_count = None
        self.pitch = None
        self.volume = None
        self.ignore_timescale = None
        self.device = None
        self.on_finish = False
        self._handle = None
        self.DONE = self.add_output(self.get_done)
        self.ON_FINISH = self.add_output(self.get_on_finish)
        self.HANDLE = self.add_output(self.get_handle)

    def get_handle(self):
        return self._handle

    def get_on_finish(self):
        if not self._handle:
            return False
        if self._handle.finished:
            return True
        return False

    def get_done(self):
        return self._done

    def reset(self):
        super().reset()
        if self._handle and self._handle.finished:
            self._handle = None

    def evaluate(self):
        self.on_finish = False
        pitch = self.get_input(self.pitch)
        volume = self.get_input(self.volume)
        if self._handle:
            self._handle.pitch = pitch
            self._handle.volume = volume
        if not self.get_condition():
            return
        file: Sound = self.get_input(self.sound)
        loop_count = self.get_input(self.loop_count)
        ignore_timescale = self.get_input(self.ignore_timescale)
        self._handle = Sound2D(
            file.filepath,
            volume,
            pitch,
            loop_count,
            False,
            ignore_timescale,
            'ln_audio_system'
        )
        self._done = True



from bge import logic
from bge.types import KX_GameObject as GameObject
from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.input.gamepad import GamepadLook
from math import degrees


class ULGamepadLook(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.axis: int = 0
        self.condition = None
        self.main_obj: GameObject = None
        self.head_obj: GameObject = None
        self.inverted: bool = None
        self.index: int = None
        self.sensitivity: float = None
        self.exponent: float = None
        self.use_cap_x: bool = None
        self.cap_x: Vector = None
        self.use_cap_y: bool = None
        self.cap_y: Vector = None
        self.threshold: float = None
        self.gamepadlook: GamepadLook = None
        self._done: bool = None
        self.DONE = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        condition = self.get_condition()
        if condition:
            obj = self.get_input(self.main_obj)
            head = self.get_input(self.head_obj)
            if self.gamepadlook and self.gamepadlook.active:
                if (
                    self.gamepadlook.obj is not obj
                    or
                    self.gamepadlook.head is not head
                ):
                    self.gamepadlook.disable()
                    self.gamepadlook = None
            if self.gamepadlook is None:
                use_cap_x = self.get_input(self.use_cap_x)
                use_cap_y = self.get_input(self.use_cap_y)
                cap_x = [degrees(x) for x in self.get_input(self.cap_x)] if use_cap_x else None
                cap_y = [degrees(y) for y in self.get_input(self.cap_y)] if use_cap_y else None
                self.gamepadlook = GamepadLook(
                    obj=obj,
                    head=head,
                    sensitivity=self.get_input(self.sensitivity),
                    use_cap_x=use_cap_x,
                    cap_x=cap_x,
                    use_cap_y=use_cap_y,
                    cap_y=cap_y,
                    invert=self.get_input(self.inverted),
                    smoothing=0.0,
                    local=head is not obj,
                    front=self.get_input(self.axis),
                    idx=self.get_input(self.index),
                    stick='RS' if self.axis else 'LS',
                    threshold=self.get_input(self.threshold),
                    exponent=self.get_input(self.exponent)
                )
        if self.gamepadlook is not None:
            self.gamepadlook.active = condition
        self._done = True


from uplogic.nodes import ULParameterNode


class ULGetFont(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.font = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self.get_input(self.font)



import sys, os
from uplogic.console import error
from uplogic.console import success
from uplogic.console import warning
from uplogic.console import debug
from uplogic.utils.math import get_local
from uplogic.utils.raycasting import raycast
from uplogic.utils import check_vr_session_status
from uplogic.utils.visualize import draw_arrow, draw_cube
from uplogic.events import schedule
from uplogic.input.vr import VR_HEADSET, VR_STATE
from sys import platform
import bge, bpy
from bge.types import KX_GameObject
from mathutils import Vector, Matrix

try:
    import pyfmodex
except ImportError:
    error('"pyfmodex" module missing, please install!')

pypath = sys.executable
if platform == "linux" or platform == "linux2":
    error('FMod: Linux support not added yet, please consider contributing.')
elif platform == "darwin":
    error('FMod: OS X not supported, please consider contributing.')
elif platform == "win32":
    debug("Loading FMod...")
    version = bpy.app.version
    dllpath = os.path.join(os.getcwd(), f'{version[0]}.{version[1]}', 'python', 'DLLs')
    fmodstudioL = os.path.join(dllpath, "fmodstudioL.dll")
    fmodL = os.path.join(dllpath, "fmodL.dll")
    exists_fmodstudioL = os.path.exists(fmodstudioL)
    exists_fmodL = os.path.exists(fmodL)
    if not (exists_fmodL and exists_fmodstudioL):
        if not exists_fmodstudioL:
            error(f'Missing "{fmodstudioL}"')
        if not exists_fmodL:
            error(f'Missing "{fmodL}"')
        error('One or more FMod Libraries not found, go to "https://www.fmod.com/download" and install FMOD Engine, then from ".../api/core/lib/x64" copy "fmodL.dll" and from ".../api/studio/lib/x64" copy "fmodstudioL.dll" to your local "python/dlls" installation.')
        sys.exit(0)
    else:
        os.environ["PYFMODEX_STUDIO_DLL_PATH"] = fmodstudioL
        os.environ["PYFMODEX_DLL_PATH"] = fmodL
        success('FMod libraries successfully loaded. Please check license at "https://www.fmod.com/licensing".')
        success(f'Using pyfmodex version {pyfmodex.__version__}.')




from pyfmodex import studio as fstudio
from pyfmodex.studio import enums
from pyfmodex import flags


version = pyfmodex.__version__.split('.')
if int(version[1]) < 7 and int(version[2]) < 2:
    error(f'"pyfmodex" module version {version} not supported, please update!')
    sys.exit(0)


def get_studio():
    FMod.initialize()
    return FMod.studio


class Sound:
    _mode = flags.MODE.TWOD

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, val):
        self._position = Vector(val)

    @property
    def velocity(self):
        return self._velocity

    @velocity.setter
    def velocity(self, val):
        self._velocity = Vector(val)
    
    @property
    def is_valid(self):
        return False

    @property
    def is_virtual(self):
        return False

    @property
    def paused(self):
        return True

    @property
    def occluded(self):
        return False

    def visualize(self, color=(1, 1, 1, 1), size=.2):
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError

    def update(self):
        raise NotImplementedError


class File2D(Sound):

    def __init__(self, path, channel='default') -> None:
        self.channel = FMod.channels.get(channel, None)
        self.channel.sounds.append(self)
        system = FMod.studio.core_system
        self.sound: pyfmodex.sound.Sound = system.create_sound(path, self._mode)
        self.channel: pyfmodex.channel.Channel = system.play_sound(self.sound, paused=True)
        self.channel.paused = False

    def update(self):
        pass

    def stop(self):
        self.channel.stop()
        self.sound.release()
        self.channel.sounds.remove(self)


class File3D(File2D):
    _mode = flags.MODE.THREED

    def __init__(self, path, position=Vector((0, 0, 0)), channel='default') -> None:
        super().__init__(path, channel)


class Event(Sound):
    
    def __init__(self, name, position=Vector((0, 0, 0)), channel='default') -> None:
        self._orientation = Matrix()
        self.channel = FMod.channels.get(channel, None)
        self.evt = FMod.studio.get_event(name).create_instance()
        print('START')
        self.occlusion_mask = self.channel.occlusion_mask
        self.evt.start()
        self.velocity = Vector((0, 0, 0))
        FMod.studio.update()
        self._position = position
        self.channel.sounds.append(self)
        self.occlusion_near_clipping = 0.0
        self._caster = bge.logic.getCurrentScene().active_camera

    @property
    def ray_caster(self):
        return self._caster

    @ray_caster.setter
    def ray_caster(self, val):
        self._caster = val

    @property
    def is_valid(self):
        return self.evt.is_valid

    @property
    def is_virtual(self):
        return self.evt.is_virtual

    @property
    def channel_group(self):
        return self.evt.channel_group

    @property
    def channel_group(self):
        return self.evt.channel_group

    @property
    def volume(self):
        return self.evt.get_volume()

    @volume.setter
    def volume(self, vol):
        self.evt.set_volume(vol)

    @property
    def pitch(self):
        return self.evt.get_pitch()

    @pitch.setter
    def pitch(self, vol):
        self.evt.set_pitch(vol)

    @property
    def orientation(self):
        return self._orientation

    @orientation.setter
    def orientation(self, val):
        self.evt.forward = val @ Vector((0, 1, 0))
        self.evt.up = val @ Vector((0, 0, 1))
        self._orientation = val

    @property
    def up(self):
        return self.orientation @ Vector((0, 0, 1))

    @property
    def forward(self):
        return self.orientation @ Vector((0, 1, 0))

    @property
    def occluded(self):
        direction = (FMod.listener.worldPosition - self.position).normalized()
        ray = raycast(
            self._caster,
            self.position + direction * self.occlusion_near_clipping,
            FMod.listener.worldPosition,
            mask=self.occlusion_mask,
            visualize=True
        )
        while ray.obj and not ray.obj.blenderObject.get('sound_occluder', True):
            ray = raycast(
                ray.obj,
                ray.point,
                FMod.listener.worldPosition,
                mask=self.occlusion_mask,
                visualize=True
            )
        return ray.obj is not None
            

    @property
    def paused(self):
        return self.evt.paused

    @paused.setter
    def paused(self, val):
        self.evt.paused = val

    @property
    def playback_state(self):
        return self.evt.playback_state

    @property
    def timeline_position(self):
        return self.evt.timeline_position

    def update(self):
        if self.evt.playback_state is enums.PLAYBACK_STATE.STOPPED:
            self.evt.release()
            self.stop()
            return
        cam = bge.logic.getCurrentScene().active_camera
        self.evt.set_3d_attributes(get_local(cam, self.position), self.velocity, self.forward)

    def visualize(self, color=(1, 1, 1, 1), size=.2):
        draw_arrow(self.position, self.position + Vector(self.evt.forward) * size, color)
        # draw_cube(self.position, size * .5, centered=True)

    def set_parameter(self, parameter, value, ignore_seek_speed=False):
        self.evt.set_parameter_by_name(parameter, value, ignore_seek_speed)

    def get_parameter(self, parameter, actual=False):
        param = self.evt.get_parameter_by_name(parameter) 
        return param[1] if actual else param[0]

    def stop(self):
        self.evt.stop()
        if self in self.channel.sounds:
            self.channel.sounds.remove(self)
        self.evt.release()


class EventSpeaker(Event):

    def __init__(self, name, speaker: KX_GameObject, channel='default') -> None:
        self.speaker = speaker
        super().__init__(name, speaker.worldPosition, channel)
        self._caster = speaker

    @property
    def velocity(self):
        return self.speaker.worldLinearVelocity * self._velocity

    @velocity.setter
    def velocity(self, val):
        self._velocity = Vector(val)

    @property
    def position(self):
        return self.speaker.worldPosition

    @position.setter
    def position(self, val):
        self.speaker.worldPosition = Vector(val)

    @property
    def orientation(self):
        return self.speaker.worldOrientation

    @orientation.setter
    def orientation(self, val):
        self.evt.forward = val @ Vector((0, 1, 0))
        self.evt.up = val @ Vector((0, 0, 1))
        self.speaker.worldOrientation = val


class Channel(dict):
    def __init__(self, name, occlusion_mask=65535):
        self.name = name
        self.sounds: list[Sound] = []
        self.occlusion_mask = occlusion_mask

    @property
    def occlusion_mask(self):
        return self._occlusion_mask

    @occlusion_mask.setter
    def occlusion_mask(self, val):
        self._occlusion_mask = val
        for s in self.sounds:
            s.occlusion_mask = val

    def set(self, key, value):
        self[key] = value
        for sound in self.sounds:
            sound.evt.set_parameter_by_name(key, value)
    
    def destroy(self):
        for sound in self.sounds.copy():
            sound.stop()
        del FMod.channels[self.name]
    
    def event(self, event, position):
        evt = Event(event, position, self.name)
        return evt

    def update(self):
        for sound in self.sounds:
            sound.update()


class FMod:
    channels = {'default': Channel('default')}
    studio: fstudio.StudioSystem = None
    listener = None

    @classmethod
    def initialize(cls):
        if cls.studio is None:
            fmodstudio = fstudio.StudioSystem()
            fmodstudio.initialize()
            cls.studio = fmodstudio
            scene = bge.logic.getCurrentScene()
            cls.listener = VR_HEADSET if VR_STATE else scene.active_camera
            scene.pre_draw.append(cls.update)
            scene.onRemove.append(cls.destroy)
        return cls.studio

    @classmethod
    def set_occlusion_mask(self, mask: int):
        for channel in self.channels.values():
            channel.occlusion_mask = mask

    @classmethod
    def load_bank(cls, path):
        if not os.path.exists(path):
            error(f"Couldn't load bank from '{path}'")
        if cls.studio is not None:
            from pyfmodex.exceptions import FmodError
            try:
                cls.studio.load_bank_file(path)
            except FmodError:
                warning(f"Bank '{path}' already loaded!")
                return
        success(f"Bank '{path}' successfully loaded.")

    @classmethod
    def update(cls):
        
        scene = bge.logic.getCurrentScene()
        cls.listener = VR_HEADSET if VR_STATE else scene.active_camera
        studio = cls.studio
        cam = cls.listener
        studio.core_system.listener().position = cam.worldPosition
        studio.core_system.listener().set_orientation(
            list(cam.getAxisVect((0, 0, 1))),
            list(cam.getAxisVect((0, 1, 0)))
        )
        studio.update()
        for channel in cls.channels.values():
            channel.update()

    @classmethod
    def add_channel(cls, name):
        cls.channels[name] = Channel(name)

    @classmethod
    def set_channel_parameter(cls, parameter_name, value, channel='default'):
        if cls.studio is None:
            return
        _channel = cls.channels.get(channel, None)
        if _channel is None:
            error(f'Channel {channel} not found.')
            return
        _channel.set(parameter_name, value)

    @classmethod
    def set_channel_occlusion_mask(cls, mask=65535, channel='default'):
        _channel = cls.channels.get(channel, None)
        if _channel is None:
            error(f'Channel {channel} not found.')
            return
        _channel.occlusion_mask = mask

    @classmethod
    def destroy(cls):
        scene = bge.logic.getCurrentScene()
        if cls.update in scene.pre_draw:
            scene.pre_draw.remove(cls.update)
            cls.studio.release()
            cls.studio = None

    @classmethod
    def event(cls, event, source=Vector((0, 0, 0)), channel='default') -> Event | EventSpeaker:
        if cls.studio is None:
            return
        _channel = cls.channels.get(channel, None)
        if _channel is None:
            _channel = cls.channels[channel] = Channel(channel)
        if isinstance(source, KX_GameObject):
            evt = EventSpeaker(f'event:/{event}', source, channel)
        else:
            evt = Event(f'event:/{event}', Vector(source), channel)
        return evt

    @classmethod
    def file3d(cls, path, channel='default'):
        if cls.studio is None:
            return
        _channel = cls.channels.get(channel, None)
        if _channel is None:
            _channel = cls.channels[channel] = Channel(channel)
        return File3D(path, channel)


FMod.initialize()


def load_bank(path):
    FMod.load_bank(path)


def start_event(event, source=Vector((0, 0, 0)), channel='default'):
    return FMod.event(event, source, channel)


def set_occlusion_mask(mask):
    FMod.set_occlusion_mask(mask)


def set_channel_occlusion_mask(mask):
    FMod.set_channel_occlusion_mask(mask)


def set_channel_parameter(parameter, value, channel='default'):
    FMod.set_channel_parameter(parameter, value, channel)


def add_channel(name):
    FMod.add_channel(name)


from uplogic.nodes import ULActionNode


class ULSetBoneConstraintInfluence(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.armature = None
        self.bone = None
        self.constraint = None
        self.influence = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        armature = self.get_input(self.armature)
        bone = self.get_input(self.bone)
        constraint = self.get_input(self.constraint)
        influence = self.get_input(self.influence)
        (
            armature
            .blenderObject
            .pose
            .bones[bone]
            .constraints[constraint]
            .influence
        ) = influence
        self._done = True



from uplogic.nodes import ULConditionNode
from uplogic.input import key_tap
from uplogic.input import key_down
from uplogic.input import key_up


class ULKeyPressed(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.pulse = False
        self.input_type = 1 if self.pulse else 0
        self.key_code = 0
        self.network = None
        self.OUT = self.add_output(self.get_pressed)

    def get_pressed(self):
        keycode = self.get_input(self.key_code)
        funcs = [key_tap, key_down, key_up]
        return funcs[self.input_type](keycode)



from uplogic.nodes import ULParameterNode


class ULGetUIWidgetAttr(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.widget = None
        self.widget_attr = 'show'
        self.BOOL = self.add_output(self._get_attr)
        self.COLOR = self.add_output(self._get_attr)
        self.ALPHA = self.add_output(self._get_attr)
        self.VEC2 = self.add_output(self._get_attr)
        self.STR = self.add_output(self._get_attr)
        self.INT = self.add_output(self._get_attr)
        self.FLOAT = self.add_output(self._get_attr)
        self.FONT = self.add_output(self._get_attr)
        self.IMG = self.add_output(self._get_attr)
        self.POINTS = self.add_output(self._get_attr)

    def _get_attr(self):
        widget = self.get_input(self.widget)
        return getattr(widget, self.widget_attr, None)



from mathutils import Vector
from uplogic.nodes import ULParameterNode


class ULVectorXYZ(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.input_x = None
        self.input_y = None
        self.input_z = None
        self.OUTV = self.add_output(self.get_out_v)

    def get_out_v(self):
        return Vector((
            self.get_input(self.input_x),
            self.get_input(self.input_y),
            self.get_input(self.input_z)
        ))




from .createuicanvas import ULCreateUICanvas
from .createuilayout import ULCreateUILayout
from .createuibutton import ULCreateUIButton
from .createuilabel import ULCreateUILabel
from .createuiimage import ULCreateUIImage
from .createuislider import ULCreateUISlider
from .createuipath import CreateUIPathNode
from .movewidget import MoveWidgetNode


from uplogic.nodes import ULParameterNode


class StringOperationNode(ULParameterNode):
    
    
    def __init__(self):
        ULParameterNode.__init__(self)
        self.string_a = ''
        self.string_b = ''
        self.string_c = ''
        self.zfill_width = 0
        self.operator = 0
        self.RESULT = self.add_output(self.get_result)
        self.operations = [
            self.get_join,
            self.get_split,
            self.get_contains,
            self.get_count,
            self.get_replace,
            self.get_startswith,
            self.get_endswith,
            self.get_upper,
            self.get_lower,
            self.get_zfill
        ]

    def get_result(self):
        return self.operations[self.operator]()

    def get_join(self):
        return str(self.get_input(self.string_a)) + str(self.get_input(self.string_b))

    def get_split(self):
        return str(self.get_input(self.string_a)).split(
            str(self.get_input(self.string_b))
        )

    def get_contains(self):
        return str(self.get_input(self.string_b)) in str(self.get_input(self.string_a))

    def get_count(self):
        return str(self.get_input(self.string_a)).count(
            str(self.get_input(self.string_b))
        )

    def get_replace(self):
        return str(self.get_input(self.string_a)).replace(
            str(self.get_input(self.string_b)),
            str(self.get_input(self.string_c))
        )

    def get_startswith(self):
        return str(self.get_input(self.string_a)).startswith(
            str(self.get_input(self.string_b))
        )

    def get_endswith(self):
        return str(self.get_input(self.string_a)).endswith(
            str(self.get_input(self.string_b))
        )

    def get_upper(self):
        return str(self.get_input(self.string_a)).upper()

    def get_lower(self):
        return str(self.get_input(self.string_a)).lower()

    def get_zfill(self):
        return str(self.get_input(self.string_a)).zfill(
            self.get_input(self.zfill_width)
        )



from bge import logic
from uplogic.nodes import ULParameterNode


class ULActiveCamera(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.OUT = self.add_output(self.get_camera)

    def get_camera(self):
        scene = logic.getCurrentScene()
        return scene.active_camera



from uplogic.nodes import ULActionNode
from bge.types import KX_GameObject


class ULSetLightColor(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.lamp = None
        self.color = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        light: KX_GameObject = self.get_input(self.lamp)
        color = self.get_input(self.color)
        if len(color) > 3:
            color = color[:-1]
        light = light.blenderObject.data
        light.color = color
        self._done = True



from uplogic.nodes import ULActionNode


class ULSetPyInstanceAttr(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.instance = None
        self.attr = None
        self.value = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        instance = self.get_input(self.instance)
        attr = self.get_input(self.attr)
        value = self.get_input(self.value)
        setattr(instance, attr, value)
        self._done = True



from uplogic.nodes import ULActionNode
from bge import logic
from bpy.types import Scene


class ULSetScene(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.scene = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        scene: Scene = self.get_input(self.scene)
        logic.getCurrentScene().replace(scene.name)
        self._done = True



# from .rendertoscreen import RenderToScreen
from bge.types import KX_Camera
from bge import render, logic
from uplogic import console


class SplitScreen:
    def __init__(self, camera2: KX_Camera = None, camera3: KX_Camera=None, camera4: KX_Camera=None) -> None:
        self.cameras: list[KX_Camera] = []
        cam = self._camera = logic.getCurrentScene().active_camera
        self.add_camera(cam)
        if camera2 is not None:
            self.add_camera(camera2)
        if camera3 is not None:
            self.add_camera(camera3)
        if camera4 is not None:
            self.add_camera(camera4)

    def disable(self):
        for cam in self.cameras:
            cam.useViewport = False
        self._camera.useViewport = False

    def enable(self):
        for cam in self.cameras:
            cam.useViewport = True
        self._arrange()

    def add_camera(self, camera: KX_Camera):
        cameras = self.cameras
        if len(cameras) == 4:
            console.warning('Maximum amount of cameras reached!')
            return
        cameras.append(camera)
        if len(cameras) == 0:
            return
        self._arrange()

    def remove_camera(self, idx=-1):
        if len(self.cameras) == 0:
            return
        cam = self.cameras.pop(idx)
        cam.setViewport(
            -1,
            -1,
            -1,
            -1
        )
        self._arrange()

    def _arrange(self):
        cameras = self.cameras
        cam_amount = len(cameras)
        width = render.getWindowWidth()
        height = render.getWindowHeight()
        offset_x = 0
        offset_y = 0
        # self._path_x.show = False
        # self._path_y.show = False
        if cam_amount > 1:
            # self._path_y.show = True
            width = int(width * .5)
            offset_x = width
        if cam_amount > 2:
            # self._path_x.show = True
            height = int(height * .5)
            offset_y = height

        cameras[0].setViewport(
            0,
            offset_y + 1,
            width,
            offset_y + height
        )
        cameras[0].useViewport = True
        if cam_amount > 1:
            cameras[1].setViewport(
                offset_x + 1,
                offset_y + 1,
                offset_x + width,
                offset_y + height
            )
            cameras[1].useViewport = True
        if cam_amount > 2:
            cameras[2].setViewport(
                0,
                0,
                width,
                height
            )
            cameras[2].useViewport = True
        if cam_amount > 3:
            cameras[3].setViewport(
                offset_x + 1,
                0,
                offset_x + width,
                height
            )
            cameras[3].useViewport = True



from bge import logic
from uplogic.nodes import ULActionNode
from bpy.types import Collection
from bge.types import KX_Camera

class ULSetOverlayCollection(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.camera = None
        self.collection = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        collection:Collection = self.get_input(self.collection)
        camera: KX_Camera = self.get_input(self.camera)
        logic.getCurrentScene().addOverlayCollection(camera, collection)
        self._done = True



from .cursorbehavior import ULCursorBehavior  # noqa
from .cursorvisibility import ULSetCursorVisibility  # noqa
from .endgame import ULEndGame  # noqa
from .gamepadvibration import ULGamepadVibration  # noqa
from .instream import ULKeyLogger  # noqa
from .loadblendfile import ULLoadBlendFile  # noqa
from .loadgame import ULLoadGame  # noqa
from .mouselook import ULMouseLook  # noqa
from .mousesetposition import ULSetMousePosition  # noqa
from .removeoverlaycollection import ULRemoveOverlayCollection  # noqa
from .restartgame import ULRestartGame  # noqa
from .runactuator import ULRunActuator  # noqa
from .runpython import ULRunPython  # noqa
from .savegame import ULSaveGame  # noqa
from .setactuatorvalue import ULSetActuatorValue  # noqa
from .setoverlaycollection import ULSetOverlayCollection  # noqa
from .setpyinstanceattr import ULSetPyInstanceAttr  # noqa
from .setsensorvalue import ULSetSensorValue  # noqa
from .startsound2d import ULStartSound  # noqa
from .startsound import StartSoundNode  # noqa
from .startsound3d import ULStartSound3D  # noqa
from .startspeaker import ULStartSpeaker  # noqa
from .dispatchevent import ULDispatchEvent  # noqa
from .addobject import ULAddObject  # noqa
from .setproperty import ULSetProperty  # noqa
from .setmaterial import ULSetMaterial  # noqa
from .setmatnodevalue import ULSetMatNodeValue  # noqa
from .playsequence import ULPaySequence  # noqa
from .setnodesocket import ULSetNodeSocket  # noqa
from .setnodevalue import ULSetNodeValue  # noqa
from .toggleproperty import ULToggleProperty  # noqa
from .setmatnodesocket import ULSetMatNodeSocket  # noqa
from .copyproperty import ULCopyProperty  # noqa
from .modifyproperty import ULModifyProperty  # noqa
from .clampedmodifyproperty import ULClampedModifyProperty  # noqa
from .printvalue import ULPrintValue  # noqa
from .createvehicle import ULCreateVehicle  # noqa
from .vehicleapplyforce import ULVehicleApplyForce  # noqa
from .vehicleapplybraking import ULVehicleApplyBraking  # noqa
from .vehicleapplysteering import ULVehicleApplySteering  # noqa
from .vehiclesetattributes import ULVehicleSetAttributes  # noqa
from .setgameobjectattr import ULSetGameObjectAttribue  # noqa
from .sendmessage import ULSendMessage  # noqa
from .setvisibility import ULSetVisibility  # noqa
from .setvisibility import ULSetCollectionVisibility  # noqa
from .setcurvepoints import ULSetCurvePoints  # noqa
from .raycast import ULRaycast  # noqa
from .raycast import RaycastNode  # noqa
from .projectileraycast import ULProjectileRayCast  # noqa
from .mouseraycast import ULMouseRayCast  # noqa
from .cameraraycast import ULCameraRayCast  # noqa
from .setcamera import ULSetCamera  # noqa

from .addphysicsconstraint import ULAddPhysicsConstraint  # noqa
from .alignaxistovector import ULAlignAxisToVector  # noqa
from .appendlistitem import ULAppendListItem  # noqa
from .applyforce import ULApplyForce  # noqa
from .applyimpulse import ULApplyImpulse  # noqa
from .applymovement import ULApplyMovement  # noqa
from .applyrotation import ULApplyRotation  # noqa
from .applytorque import ULApplyTorque  # noqa
from .characterjump import ULCharacterJump  # noqa
from .clearvariables import ULClearVariables  # noqa
from .drawbox import ULDrawBox  # noqa
from .drawcube import ULDrawCube  # noqa
from .drawline import ULDrawLine  # noqa
from .editbone import ULEditBone  # noqa
from .endobject import ULEndObject  # noqa
from .executesubnetwork import ULExecuteSubNetwork  # noqa
from .followpath import ULFollowPath  # noqa
from .gamepadlook import ULGamepadLook  # noqa
from .getperformanceprofile import ULGetPerformanceProfile  # noqa
from .installsubnetwork import ULInstallSubNetwork  # noqa
from .listglobalvalues import ULListGlobalValues  # noqa
from .listvariables import ULListVariables  # noqa
from .makeuniquelight import ULMakeUniqueLight  # noqa
from .moveto import ULMoveTo  # noqa
from .movetowithnavmesh import ULMoveToWithNavmesh  # noqa
from .pausesound import ULPauseSound  # noqa
from .playaction import ULPlayAction  # noqa
from .popdictkey import ULPopDictKey  # noqa
from .removelistindex import ULRemoveListIndex  # noqa
from .removelistvalue import ULRemoveListValue  # noqa
from .removeparent import ULRemoveParent  # noqa
from .removephysicsconstraint import ULRemovePhysicsConstraint  # noqa
from .removevariable import ULRemoveVariable  # noqa
from .replacemesh import ULReplaceMesh  # noqa
from .resumesound import ULResumeSound  # noqa
# XXX. Rename Class to ULRotateTo
from .rotateto import ULActionRotateTo  # noqa
from .savevariable import ULSaveVariable  # noqa
from .savevariabledict import ULSaveVariableDict  # noqa
from .setactionframe import ULSetActionFrame  # noqa
from .setboneconstraintattr import ULSetBoneConstraintAttribute  # noqa
from .setboneconstraintinfluence import ULSetBoneConstraintInfluence  # noqa
from .setboneconstrainttarget import ULSetBoneConstraintTarget  # noqa
from .setboneposition import ULSetBonePosition  # noqa
from .setcamerafov import ULSetCameraFOV  # noqa
from .setcameraorthoscale import ULSetCameraOrthoScale  # noqa
from .setcharactergravity import ULSetCharacterGravity  # noqa
from .setcharacterjumpspeed import ULSetCharacterJumpSpeed  # noqa
from .setcharactermaxjumps import ULSetCharacterMaxJumps  # noqa
from .setcharactervelocity import ULSetCharacterVelocity  # noqa
from .setcharacterwalkdir import ULSetCharacterWalkDir  # noqa
from .setcollisiongroup import ULSetCollisionGroup  # noqa
from .setcollisionmask import ULSetCollisionMask  # noqa
from .setdictkey import ULSetDictKey  # noqa
from .setdynamics import ULSetDynamics  # noqa
from .seteeveeao import ULSetEeveeAO  # noqa
from .seteeveebloom import ULSetEeveeBloom  # noqa
from .seteeveesmaa import ULSetEeveeSMAA  # noqa
from .seteeveesmaaquality import ULSetEeveeSMAAQuality  # noqa
from .seteeveessr import ULSetEeveeSSR  # noqa
from .seteeveevolumetrics import ULSetEeveeVolumetrics  # noqa
from .setexposure import ULSetExposure  # noqa
from .setfullscreen import ULSetFullscreen  # noqa
from .setgamma import ULSetGamma  # noqa
from .setglobalvalue import ULSetGlobalValue  # noqa
from .setgravity import ULSetGravity  # noqa
from .setlightcolor import ULSetLightColor  # noqa
from .setlightenergy import ULSetLightEnergy  # noqa
from .setlightshadow import ULSetLightShadow  # noqa
from .setlistindex import ULSetListIndex  # noqa
from .setparent import ULSetParent  # noqa
from .setphysics import ULSetPhysics  # noqa
from .setprofile import ULSetProfile  # noqa
from .setresolution import ULSetResolution  # noqa
from .setrigidbody import ULSetRigidBody  # noqa
from .settimescale import ULSetTimeScale  # noqa
from .setvsync import ULSetVSync  # noqa
from .showframerate import ULShowFramerate  # noqa
from .startsubnetwork import ULStartSubNetwork  # noqa
from .stopaction import ULStopAction  # noqa
from .stopallsounds import ULStopAllSounds  # noqa
from .stopsound import ULStopSound  # noqa
from .stopsubnetwork import ULStopSubNetwork  # noqa
from .translate import ULTranslate  # noqa
from .addfilter import ULAddFilter  # noqa
from .removefilter import ULRemoveFilter  # noqa
from .togglefilter import ULToggleFilter  # noqa
from .setfilterstate import ULSetFilterState  # noqa
from .setscene import ULSetScene  # noqa
from .loadscene import ULLoadScene  # noqa
from .loadfilecontent import ULLoadFileContent  # noqa
from .slowfollow import ULSlowFollow  # noqa
from .setcustomcursor import ULSetCustomCursor  # noqa
from .adduiwidget import ULAddUIWidget
from .setuiwidgetattr import ULSetUIWidgetAttr
from .spawnpool import ULSpawnPool
from .localserver import ULLocalServer
from .localclient import ULLocalClient
from .sendnetworkmessage import ULSendNetworkMessage
from .applytransform import ULApplyTransform
from .setlogictreeproperty import ULSetLogicTreeProperty
from .draw import DrawNode
from .setconstraintattribute import SetConstraintAttributeNode
from .setrigboneattribute import SetRigBoneAttributeNode
from .setportal import SetPortalNode
from .storevalue import ULStoreValue
from .distributecurvepoints import DistributeCurvePointsNode
from .togglelogictreeproperty import ToggleLogicTreePropertyNode



from uplogic.nodes import ULActionNode
from uplogic.utils import logic
from bpy.types import Collection
from bge.types import KX_GameObject
from bge.types import KX_Scene
import bpy


class ULSetVisibility(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.visible: bool = None
        self.recursive: bool = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_object: KX_GameObject = self.get_input(self.game_object)
        visible: bool = self.get_input(self.visible)
        recursive: bool = self.get_input(self.recursive)
        game_object.setVisible(visible, recursive)
        self._done = True


class ULSetCollectionVisibility(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.collection = None
        self.visible: bool = None
        self.recursive: bool = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def set_collection_visible(
        self,
        visible,
        recursive,
        scene: KX_Scene,
        collection:Collection
    ):
        for o in collection.objects:
            gameObject = scene.getGameObjectFromObject(o)
            if gameObject:
                gameObject.setVisible(visible, True)
        if recursive:
            if len(collection.children) <= 0:
                return
            for child in collection.children:
                self.set_collection_visible(visible, recursive, scene, child)

    def evaluate(self):
        if not self.get_condition():
            return
        collection: Collection = self.get_input(self.collection)
        visible: bool = self.get_input(self.visible)
        recursive: bool = self.get_input(self.recursive)
        scene = logic.getCurrentScene()
        self.set_collection_visible(visible, recursive, scene, collection)
        self._done = True



import operator
from mathutils import Vector
from bge import logic
import math
from .math import matmul
from .math import multadd
from .math import isqrt
from .math import smin
from .math import smax
from .math import sign
from .math import compare
from .math import fraction
from .math import trunc_mod
from .math import floor_mod
from .math import wrap
from .math import snap
from .math import ping_pong
from .math import _round
from .math import _log
from .math import _min
from .math import _max
from .math import _asin
from .math import _acos
from .math import _atan
from .math import _lerp
import bpy


# uplogic game properties
VEHICLE = '.ulvehicleconst'
SHIP = '.ulshipconst'
FLOTSAM = '.ulflotsamconst'
WATER = '.ulwater'

STREAMTYPE_DOWNSTREAM = 0
STREAMTYPE_UPSTREAM = 1

DISCONNECT_MSG = '!DISCONNECT'


LOGIC_OPERATORS = [
    operator.eq,
    operator.ne,
    operator.gt,
    operator.lt,
    operator.ge,
    operator.le
]

OPERATORS = {
    "ADD": operator.add,
    "DIV": operator.truediv,
    "MUL": operator.mul,
    "SUB": operator.sub,
    'FDIV': operator.floordiv,
    'MATMUL': matmul, #operator.matmul,
    'MOD': operator.mod,
    'POW': operator.pow
}

MATH_OPERATORS = [
    operator.add,  # 0
    operator.sub,  # 1
    operator.mul,  # 2
    operator.truediv,  # 3
    multadd,  # 4
    operator.mod,  # 5
    operator.floordiv,  # 6
    operator.pow,  # 7
    _log,  # 8
    math.sqrt,  # 9
    isqrt,  # 10
    abs,  # 11
    math.exp,  # 12
    _min,  # 13
    _max,  # 14
    operator.lt,  # 15
    operator.gt,  # 16
    sign,  # 17
    compare,  # 18
    smin,  # 19
    smax,  # 20
    _round,  # 21
    math.floor,  # 22
    math.ceil,  # 23
    math.trunc,  # 24
    fraction,  # 25
    trunc_mod,  # 26
    floor_mod,  # 27
    wrap,  # 28
    snap,  # 29
    ping_pong,  # 30
    math.sin,  # 31
    math.cos,  # 32
    math.tan,  # 33
    _asin,  # 34
    _acos,  # 35
    _atan,  # 36
    math.atan2,  # 37
    math.sinh,  # 38
    math.cosh,  # 39
    math.tanh,  # 40
    math.radians,  # 41
    math.degrees,  # 42
    _lerp  # 43
]


LO_AXIS_TO_STRING_CODE = {
    0: "X", 1: "Y", 2: "Z",
    3: "-X", 4: "-Y", 5: "-Z",
}


LO_AXIS_TO_VECTOR = {
    0: Vector((1, 0, 0)), 1: Vector((0, 1, 0)),
    2: Vector((0, 0, 1)), 3: Vector((-1, 0, 0)),
    4: Vector((0, -1, 0)), 5: Vector((0, 0, -1)),
}

FRONT_AXIS_VECTOR_SIGNED = {
    0: Vector((1, 0)), 1: Vector((1, 0)),
    2: Vector((0, 1)), 3: Vector((-1, 0)),
    4: Vector((-1, 0)), 5: Vector((0, -1)),
}


FRAMETIME_COMPARE = 1 / bpy.data.scenes[logic.getCurrentScene().name].render.fps


def FPS_FACTOR() -> float:
    avg = logic.getAverageFrameRate()
    return (bpy.data.scenes[logic.getCurrentScene().name].game_settings.fps / avg) if 0 < avg < 10000 else 1

def DELTA_TIME() -> float:
    avg = logic.getAverageFrameRate()
    return (1 / avg) if 0 < avg < 10000 else 0.0


RED = [1, 0, 0, 1]
GREEN = [0, 1, 0, 1]
BLUE = [0, 0, 1, 1]
YELLOW = [1, 1, 0, 1]
PURPLE = [1, 0, 1, 1]
TORQUISE = [0, 1, 1, 1]
WHITE = [1, 1, 1, 1]
BLACK = [0, 0, 0, 1]
GREY = [.5, .5, .5, 1]


class prop_type_invalid:
    ...


PROP_TYPE_INVALID = prop_type_invalid


from mathutils import Euler
from uplogic.nodes import ULParameterNode


class ULEuler(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.input_x = None
        self.input_y = None
        self.input_z = None
        self.order = 'XYZ'
        self.OUTV = self.add_output(self.get_out_v)

    def get_out_v(self):
        e = Euler(order=self.order)
        x = self.get_input(self.input_x)
        y = self.get_input(self.input_y)
        z = self.get_input(self.input_z)
        e.x = x
        e.y = y
        e.z = z
        return e



from uplogic.nodes import ULActionNode


class ULSetActuatorValue(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_obj = None
        self.act_name = None
        self.field = None
        self.value = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        game_obj = self.get_input(self.game_obj)
        act_name = self.get_input(self.act_name)
        if not self.get_condition():
            return
        actuator = game_obj.actuators.get(act_name)
        if not actuator:
            return
        field = self.get_input(self.field)
        value = self.get_input(self.value)
        setattr(actuator, field, value)
        self._done = True



from mathutils import Vector
from mathutils import Matrix
from bge.types import KX_GameObject


class Vec2(list):

    def __init__(self, vec2: Vector):
        self.append(vec2.x)
        self.append(vec2.y)


class Vec3(list):

    def __init__(self, vec3: Vector):
        self.append(vec3.x)
        self.append(vec3.y)
        self.append(vec3.z)


class Vec4(list):

    def __init__(self, vec4: Vector):
        self.append(vec4.x)
        self.append(vec4.y)
        self.append(vec4.z)
        self.append(vec4.w)


class Mat3(list):

    def __init__(self, mat3: Matrix):
        self.append([0, 0, 0])
        self.append([0, 0, 0])
        self.append([0, 0, 0])

        self[0][0] = mat3[0][0]
        self[0][1] = mat3[0][1]
        self[0][2] = mat3[0][2]
        self[1][0] = mat3[1][0]
        self[1][1] = mat3[1][1]
        self[1][2] = mat3[1][2]
        self[2][0] = mat3[2][0]
        self[2][1] = mat3[2][1]
        self[2][2] = mat3[2][2]


class Mat4(list):

    def __init__(self, mat4: Matrix):
        self.append([0, 0, 0, 0])
        self.append([0, 0, 0, 0])
        self.append([0, 0, 0, 0])
        self.append([0, 0, 0, 0])

        self[0][0] = mat4[0][0]
        self[0][1] = mat4[0][1]
        self[0][2] = mat4[0][2]
        self[0][3] = mat4[0][3]

        self[1][0] = mat4[1][0]
        self[1][1] = mat4[1][1]
        self[1][2] = mat4[1][2]
        self[1][3] = mat4[1][3]

        self[2][0] = mat4[2][0]
        self[2][1] = mat4[2][1]
        self[2][2] = mat4[2][2]
        self[2][3] = mat4[2][3]

        self[3][0] = mat4[3][0]
        self[3][1] = mat4[3][1]
        self[3][2] = mat4[3][2]
        self[3][3] = mat4[3][3]


class GameObj(dict):

    def __init__(self, game_obj: KX_GameObject):
        self['name'] = game_obj.name
        self['data_id'] = game_obj.blenderObject.data.name

        # self['worldPosition'] = Vec3(game_obj.worldPosition)
        # self['worldOrientation'] = Mat3(game_obj.worldOrientation)
        # self['worldScale'] = Vec3(game_obj.worldScale)
        # self['worldLinearVelocity'] = Vec3(game_obj.worldLinearVelocity)
        # self['worldAngularVelocity'] = Vec3(game_obj.worldAngularVelocity)
        self['worldTransform'] = Mat4(game_obj.worldTransform)

        # self['localPosition'] = Vec3(game_obj.localPosition)
        # self['localOrientation'] = Mat3(game_obj.localOrientation)
        # self['localScale'] = Vec3(game_obj.localScale)
        # self['localLinearVelocity'] = Vec3(game_obj.localLinearVelocity)
        # self['localAngularVelocity'] = Vec3(game_obj.localAngularVelocity)
        # self['localTransform'] = Mat4(game_obj.localTransform)
        props = self['properties'] = {}
        for prop in game_obj.getPropertyNames():
            if game_obj[prop].__class__ in [int, float, bool, str]:
                props[prop] = game_obj[prop]
        


from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.utils.visualize import draw_cube


class ULDrawCube(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.color = None
        self.origin = None
        self.width = None
        self.use_volume_origin = False
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        origin = self.get_input(self.origin)
        width = self.get_input(self.width)
        color = self.get_input(self.color)
        origin = origin.copy()
        if self.use_volume_origin:
            offset = width * 0.5
            origin -= Vector((offset, offset, offset))
        draw_cube(origin, width, color)
        self._done = True



from bge.types import KX_GameObject
from bge import logic
from bpy.types import Object
import bpy
from mathutils import Vector, Matrix, Color
from .objects import GameObject
from uplogic import console


def make_unique_light(light: KX_GameObject) -> KX_GameObject:
    '''Make a copy of this light to allow for unique modifications.

    :param light: The `KX_LightObject` of which to copy the data.

    :returns: The modified `KX_LightObject`. This return value is for usability reasons and
    is the same as the original argument.
    '''
    lamp = light.blenderObject
    lamp.data = lamp.data.copy()
    return light


class Light(GameObject):
    """Wrapper for `KX_LightObject`. Can either convert a given scene light or create a new one.

    :param name: The name of the new light (only relevant if `light` argument is None).
    :param type: The type of the new light of [`"POINT"`, `"SUN"`, `"SPOT"`, `"AREA"`] (only relevant if `light` argument is None).
    :param light: The scene light to be converted. Leave at `None` to create new.
    """

    _deprecated = False

    def __init__(
        self,
        name: str = '',
        type: str = 'POINT',
        light: KX_GameObject = None
    ) -> None:
        if self._deprecated:
            console.warning('[UPLOGIC] ULLight class will be renamed to "Light" in future releases!')
        if light:
            self.game_object = make_unique_light(light)
            """The wrapped `KX_LightObject`."""
            return
        game_scene = logic.getCurrentScene()
        scene = bpy.data.scenes[game_scene.name]
        light = bpy.data.lights.new(name, type)
        light = bpy.data.objects.new(name, light)
        scene.collection.objects.link(light)
        self.game_object = game_scene.convertBlenderObject(light)
        """The wrapped `KX_LightObject`."""
        self.energy = 10

    @property
    def energy(self) -> float:
        return self.game_object.blenderObject.data.energy

    @energy.setter
    def energy(self, val: float):
        self.game_object.blenderObject.data.energy = val

    @property
    def color(self) -> Color:
        return self.game_object.blenderObject.data.color

    @color.setter
    def color(self, val: Color):
        self.game_object.blenderObject.data.color = val

    @property
    def use_shadow(self) -> bool:
        return self.game_object.blenderObject.data.use_shadow

    @use_shadow.setter
    def use_shadow(self, val: bool):
        self.game_object.blenderObject.data.use_shadow = val

    @property
    def shadow_clip_start(self) -> float:
        return self.game_object.blenderObject.data.shadow_buffer_clip_start

    @shadow_clip_start.setter
    def shadow_clip_start(self, val: float):
        self.game_object.blenderObject.data.shadow_buffer_clip_start = val

    @property
    def shadow_bias(self) -> float:
        return self.game_object.blenderObject.data.shadow_buffer_bias

    @shadow_bias.setter
    def shadow_bias(self, val: float):
        self.game_object.blenderObject.data.shadow_buffer_bias = val

    @property
    def use_custom_distance(self) -> bool:
        return self.game_object.blenderObject.data.use_custom_distance

    @use_custom_distance.setter
    def use_custom_distance(self, val: bool):
        self.game_object.blenderObject.data.use_custom_distance = val

    @property
    def distance(self) -> float:
        return self.game_object.blenderObject.data.cutoff_distance

    @distance.setter
    def distance(self, val: float):
        self.game_object.blenderObject.data.cutoff_distance = val

    @property
    def angle(self)-> float:
        return self.game_object.blenderObject.data.angle

    @angle.setter
    def angle(self, val: float):
        self.game_object.blenderObject.data.angle = val

    @property
    def spot_size(self) -> float:
        return self.game_object.blenderObject.data.spot_size

    @spot_size.setter
    def spot_size(self, val: float):
        self.game_object.blenderObject.data.spot_size = val

    @property
    def spot_blend(self) -> float:
        return self.game_object.blenderObject.data.spot_blend

    @spot_blend.setter
    def spot_blend(self, val: float):
        self.game_object.blenderObject.data.spot_blend = val

    @property
    def radius(self) -> float:
        return self.game_object.blenderObject.data.shadow_soft_size

    @radius.setter
    def radius(self, val: float):
        self.game_object.blenderObject.data.shadow_soft_size = val


class ULLight(Light):
    _deprecated = True



from uplogic.nodes import ULActionNode
from bpy.types import Material
import bpy


class ULSetMatNodeValue(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.mat_name = None
        self.node_name = None
        self.internal = None
        self.attribute = None
        self.value = None
        self.OUT = self.add_output(self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        material: Material = self.get_input(self.mat_name)
        node_name = self.get_input(self.node_name)
        attribute = self.get_input(self.attribute)
        internal = self.get_input(self.internal)
        value = self.get_input(self.value)
        target = (
            material
            .node_tree
            .nodes[node_name]
        )
        if internal:
            target = getattr(target, internal, target)
        if hasattr(target, attribute):
            setattr(target, attribute, value)
        self._done = True
